"""------------------------------------------------------------------------------------------
Bayesian Affect Control Theory
Main classes and functions
Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey
September 2013
Use for research purposes only.
Please do not re-distribute without written permission from the author
Any commerical uses strictly forbidden.
Code is provided without any guarantees.
Research sponsored by the Natural Sciences and Engineering Council of Canada (NSERC).
use python2.6
see README for details
----------------------------------------------------------------------------------------------"""

import math
import sys
import re
import numpy as NP
import itertools


#---------------------------------------------------------------------------------------------------------------------------
#filenames for the ACT databases
#---------------------------------------------------------------------------------------------------------------------------
ofname="bayesact.spudd"
#file with the transient equations
temfname="tdynamics-male.dat"
teffname="tdynamics-female.dat"

sentiment_levels=2
smin=-4.3
smax=4.3
sincr=(smax-smin)/sentiment_levels



#for discretising the sentiment range - not really used anymore
sentiment_indices=xrange(sentiment_levels+1)
sentiment_bins=[x*sincr+smin for x in list(xrange(sentiment_levels+1))]




#---------------------------------------------------------------------------------------------------------------------------
#Some predefined constants
#---------------------------------------------------------------------------------------------------------------------------

#whether the fundamental identities are statically defined or not
f_static=False
#whole model is observable  - this is implied by x and f being observable
f_observable=False


#normally this is ["e","p","a"]
epa=["e","p","a"]
#this is usually  ["a","b","c"]
sentvals=["a","b","c"]


#create names for fundamentals and transients
fvars=[]
tvars=[]
#switched these around feeb 15th used to be e first then s
for s in sentvals:
    for e in epa:
        tvars.append("T"+s+e)
        if not f_static or not (s=="a" or s=="c"):
            fvars.append("F"+s+e)



#---------------------------------------------------------------------------------------------------------------------------
#Builds some global constants describing dynamics (this is the matrix "M")
#---------------------------------------------------------------------------------------------------------------------------

#open the input file for dynamics of transients
# these are ordered as:
# Ae, Ap, Aa, Be, Bp, Ba, Oe, Op, Oa,
#for males
tdyn_male=[]
fb=file(temfname,"rU")
x=fb.readline()
while x!="":
    y=x.split()
    tvec=list(y[0])[1:]
    tvecc=[w[0] for w in filter(lambda (z): z[1]=="1", zip(tvars,tvec))]
    M=map(lambda(z): float(z),y[1:])
    tdyn_male.append([tvecc,M])
    x=fb.readline()
fb.close()
#for females
tdyn_female=[]
fb=file(teffname,"rU")
x=fb.readline()
while x!="":
    y=x.split()
    tvec=list(y[0])[1:]
    tvecc=[w[0] for w in filter(lambda (z): z[1]=="1", zip(tvars,tvec))]
    M=map(lambda(z): float(z),y[1:])
    tdyn_female.append([tvecc,M])
    x=fb.readline()
fb.close()


#---------------------------------------------------------------------------------------------------------------------------
#statically defined functions of general usage
#---------------------------------------------------------------------------------------------------------------------------
#initialises ids
#0: knows own id, not client id
#1: knows own id, knows client id is one of num_confusers possibilities
#2: knows own id, knows client id
#3: doesn't know own id, doesn't know client id
def init_id(knowledge,agent_id,client_id=[],mean_id=[],num_confusers=0):
    if knowledge==2 and not client_id == []:
        tau_init=NP.concatenate((agent_id,NP.zeros((3,1)),client_id)).transpose()
        prop_init=[1.0]
        beta_client_init=0.001  #was 0.1
        beta_agent_init=0.001
    elif knowledge==1:
        for i in range(num_confusers+1):
            tmp_id = NP.dot(NP.random.random([3,1]),8.6)-4.3
            tau_init.append(NP.concatenate((agent_id,NP.zeros((3,1)),tmp_id)))
        ci=NP.random.randint(num_confusers+1)
        tau_init[ci]=NP.concatenate((agent_id,NP.zeros((3,1)),client_id))
        prop_init=NP.dot(NP.ones((num_confusers+1,1)),1.0/(num_confusers+1)).transpose()
        beta_client_init=0.5
        beta_agent_init=0.01
    elif knowledge==3: 
        if mean_id==[]:
            tau_init=NP.concatenate((NP.zerois(3,1),NP.zeros((3,1)),NP.zeros((3,1)))).transpose()
        else:
            tau_init=NP.concatenate((NP.asarray([mean_id]).transpose(),NP.zeros((3,1)),NP.asarray([mean_id]).transpose())).transpose()
        prop_init=[1.0]
        beta_client_init=2.0
        beta_agent_init=2.0
    else:   #knowldge is 0
        if mean_id==[]:
            tau_init=NP.concatenate((agent_id,NP.zeros((3,1)),NP.zeros((3,1)))).transpose()
        else:
            tau_init=NP.concatenate((agent_id,NP.zeros((3,1)),NP.asarray([mean_id]).transpose())).transpose()
        prop_init=[1.0]
        beta_client_init=2.0
        beta_agent_init=0.01
    return (tau_init,prop_init,beta_client_init,beta_agent_init)

#reads in sentiments for gender from fbfname and returns a dictionary
def readSentiments(fbfname,gender):
    #open the input file for fundamentals of behaviours or identities
    addto_agent=0
    if gender=="female":
        addto_agent=3
    fsentiments_agent={}
    fb=file(fbfname,"rU")
    for x in set(fb):
        y=[re.sub(r"\s+","_",i.strip()) for i in x.split(",")]
        #sentiment_bins include the end-points -4.3 and 4.3 so we strip them off.
        z=["fb"+str(j) for j in NP.digitize(y[1:-1],sentiment_bins[1:-1])]
        #1,2,3 areEPA for males, 4,5,6 are EPA for females
        #for males
        fsentiments_agent[y[0]]={"elabel":z[0+addto_agent],"e":y[1+addto_agent],
                                 "plabel":z[1+addto_agent],"p":y[2+addto_agent],
                                 "alabel":z[2+addto_agent],"a":y[3+addto_agent]}
    fb.close()
    return fsentiments_agent

#returns the mean and covariance of all identities in the dictionary for gender
def getIdentityStats(fifname,gender):
    iddic=readSentiments(fifname,gender)
    meanid = NP.zeros((1,3)).flatten(1)
    covid = NP.zeros((3,3))
    N=0
    for theid in iddic:
        thevec=NP.asarray([float(iddic[theid]["e"]),float(iddic[theid]["p"]),float(iddic[theid]["a"])])
        meanid = NP.add(meanid,thevec)
        N=N+1
    meanid = meanid/N
    for theid in iddic:
        thevec=NP.add(NP.asarray([float(iddic[theid]["e"]),float(iddic[theid]["p"]),float(iddic[theid]["a"])]),NP.dot(-1.0,meanid))
        covid = NP.add(covid,NP.dot(NP.transpose([thevec]),[thevec]))
    covid = covid/(N-1)
    return (meanid,covid)

#prints all pairs of identities that are within edist of each other (euclidean distance)
def printNearbyIdentities(fifname,gender,edist):
    iddic=readSentiments(fifname,gender)
    listofids=[]
    for theid in iddic:
        thevec=NP.asarray([float(iddic[theid]["e"]),float(iddic[theid]["p"]),float(iddic[theid]["a"])])
        for theotherid in iddic:
            if not theid == theotherid:
                thedist=math.sqrt(fdist(thevec,iddic[theotherid]))
                if thedist < edist:
                    listofids.append((theid,theotherid,thedist))
    return listofids

#gets the EPA value for identity ident with gender 
def getIdentity(fifname,theid,gender):
    iddic=readSentiments(fifname,gender)
    if iddic.has_key(theid):
        return NP.asarray([float(iddic[theid]["e"]),float(iddic[theid]["p"]),float(iddic[theid]["a"])])
    else:
        return []
    


#true if mrow[0] has a variable that starts with v in it
def has_variable(mrow,v):
    return reduce(lambda xx,yy: xx or yy.startswith(v),mrow,False)

#consumes a matrix h which is like tdyn -  has h[i][0] giving the elements of of T that are to be used
#and h[i][1] giving the coefficients of M for that element
def h_fun(h,tvars,t,turn="agent"):
    #vector of zeros of the same size as h[1]
    sum=NP.zeros_like(h[0][1])
    for hh in h:
        #only do the non-behaviour transient values
        mult = 1
        for v in filter(lambda xx: not xx[1]=="b",hh[0]):
            mult *= t[tvars.index(v)]
        sum = sum + NP.multiply(mult,hh[1]) # pointwise multiply so return the whole row h[1] multiplied by t[v]
    return sum  

def swapACVar(Tvar):
    if Tvar[1]=="a":
        return Tvar[0]+"c"+Tvar[2]
    elif Tvar[1]=="c":
        return Tvar[0]+"a"+Tvar[2]
    else:
        return Tvar
    
def swapACVars(arrayTvars):
    return map(lambda x: swapACVar(x), arrayTvars)

#also make swapped versions of these with all input and output a and c swapped
#still to be done -----
def swapAC(dynVar):
    dd=[]
    for d in dynVar:
        ddd=[]
        ddd.append(swapACVars(d[0]))
        # this is a bummer to have this fixed, but much more difficult than otherwise
        # we know the order is : Ae, Ap, Aa, Be, Bp, Ba, Oe, Op, Oa,
        ddd.append(NP.concatenate((d[1][6:9],d[1][3:6],d[1][0:3])))

        dd.append(ddd)
    return dd


#Tvar is the name of a variable in tdyn
#tvars is the array of transient variable names
#fvars is the array of fundamental variable names
#tdyn[0] is an array of arrays of transient names, each one giving the transients involved in that factor in "t"
#tdyn[1] is an array of coefficients (the matrix M)
#tau is an array of tau values for the current step  (in the same order as tvars)
#fp is an array of fundamental values for the next time step (in the same order as tvars)
#sampleTvar("Tae",tvars,tdyn,[array of Tau values],[array of f_primed values]
#although this is called "sample"Tvar, it is actually a deterministic function!
def sampleTvarOld(Tvar,tvars,tdyn,tau,fp,turn):
    if turn=="agent":
        G=map(lambda xx: xx[0], tdyn)
        M=map(lambda xx: xx[1][tvars.index(Tvar)], tdyn)
    else:
        G=map(lambda xx: swapACVars(xx[0]), tdyn)
        M=map(lambda xx: xx[1][tvars.index(swapACVar(Tvar))], tdyn)
    sum=0
    for (tvarnames,ovecval) in zip(G,M):
        term=1
        for tvarname in tvarnames:
            if tvarname[1]=="b":
                fvarname="F"+tvarname[1:] #its taken from the primed fundamentals variable with the same e and abo values
                term *= fp[tvars.index(tvarname)]  #could be fp[fvarname] - so make fp a dictionary over variable names?
            else:
                term *= tau[tvars.index(tvarname)] #its taken from the transients variable from the previous step
                                                   #could be tau[tvarname] if tau is a dictionary
        term *= ovecval  #was ... why? : M[tvars.index(tvarname)]
        sum+=term
    return sum

def sampleTvar(Tvar,tvars,GM,tau,fp,turn):
    sum=0
    for (tvarnames,ovecval) in GM[turn][Tvar]:
        term=1
        for tvarname in tvarnames:
            if tvarname[1]=="b":
                fvarname="F"+tvarname[1:] #its taken from the primed fundamentals variable with the same e and abo values
                term *= fp[tvars.index(tvarname)]  #could be fp[fvarname] - so make fp a dictionary over variable names?
            else:
                term *= tau[tvars.index(tvarname)] #its taken from the transients variable from the previous step
                                                   #could be tau[tvarname] if tau is a dictionary
        term *= ovecval  #was ... why? : M[tvars.index(tvarname)]
        sum+=term
    return sum


def sampleTvarsOld(tvars,tdyn,tau,fp,turn):
    tsample=[]
    for tvar in tvars:
        tsample.append(sampleTvarOld(tvar,tvars,tdyn,tau,fp,turn))
    return tsample

def sampleTvars(tvars,GM,tau,fp,turn):
    tsample=[]
    for tvar in tvars:
        tsample.append(sampleTvar(tvar,tvars,GM,tau,fp,turn))
    return tsample

#extracts the b variables from fvars
#flag =False : gets the "not-b" variables from fvars
def getbvars(fvars,vars,flag=True):
    if flag:
        return map(lambda x: x[1], filter(lambda x: x[0][1]=="b", zip(fvars,vars)))
    else:
        return map(lambda x: x[1], filter(lambda x: not x[0][1]=="b", zip(fvars,vars)))


#jth column of H is given by h_fun(H[j],tau)
#method can be "fact" in which case b is drawn and then a and c independently
#or not, in which case the whole distribution over f is sampled
#alpha and beta must be greater than 0 and less than infinity
#these special cases would have to be treated separately
#******method="fact" does not work.  Also unchecked if beta_value is < 0 (ignore beta_value part)
def sampleFvar(fvars,tvars,H,C,isiga,isigf,tau,f,turn="agent",aab=[],observ=[],method="full"):
    #now we will insert aab into f here
    #this seems like a hack, but in fact it is only a small one
    #we are "using" the "b" slot in f to hold what the agent action is
    #and when turn=="agent", the variance on b is tiny, so forces 
    #the samples to have a Fb-value which is the same as aab
    #here, we could also replace f[0:3] with aab if aab is 6-D :
    #f=NP.concatenate((aab,f[6:9]))
    #this would happen if the agent can also set its own identity with an action
    #the isigf would also need to be modified to reflect this
    #see in the constructor the comment about this - search for AAB6D
    if turn=="agent" and not aab==[]:
        f=NP.concatenate((f[0:3],aab,f[6:9]))

    tmpH=h_fun(H[turn]["Tbe"],tvars,tau)
    for j in ["Tbp","Tba"]:
        tmpH= NP.vstack([tmpH,h_fun(H[turn][j],tvars,tau)])
    tmpC=NP.array([h_fun(C[turn],tvars,tau)]).transpose()

    theweight=0.0

    zmat= NP.zeros((3,9))
    K= NP.eye(9)-NP.vstack([zmat, tmpH, zmat]).transpose()
    
    #including beta, full version using a draw from a multivariate_normal
    #this will be slower I guess, but I can't find a way to split these into two
    #one way might be to sample without the beta term, then weight using the difference
    # from the previous fundamental
    ka=NP.dot(K.transpose(),isiga)
        
    #this is K^{-1} \Sig K^{-1}  or (K\Sig^{-1} K)^{-1}
    #this is actually (K\Sig^{-1} K)  - not sure what the previous comment line was supposd to mean - delete?
    mean_prediction = NP.dot(ka,K)
    sig_n=NP.linalg.inv(mean_prediction+isigf[turn])

    #mean value
    mu_n=NP.dot(sig_n,(NP.dot(ka,tmpC)+NP.dot(isigf[turn],NP.array([f]).transpose())))
            
    f_s=NP.random.multivariate_normal(NP.asarray(mu_n).flatten(1),sig_n).transpose()

    return (f_s,theweight)

#evaluates a sample of Fvar'=fp
def evalSampleFvar(tvars,tdyn,state,gamma2,turn,observ):
    weight=0.0
    if (not observ==[]) and turn=="client":
        fb=getbvars(fvars,state.f)
        dvo=NP.array(fb)-NP.array(observ)
        weight += normpdf(dvo,0.0,math.sqrt(gamma2))
    else:
        weight=0.0
    return weight


def normpdf(x, mean, sd):
    size=len(x)
    theivar=1.0/(float(sd)**2)
    thedet=math.pow(theivar,float(size))
    denom=( math.pow((2*NP.pi),float(size)/2) * math.pow(thedet,1.0/2) )
    num=0.0
    for xv in x:
        num = num-theivar*0.5*( float(xv) - float(mean) )**2

    return num-math.log(denom)

def mvar_normpdf(x, mean, cov):
    size=len(x)
    icov=NP.linalg.inv(cov)
    dm=NP.asarray(x-mean)
    md=NP.transpose(dm)
    num=-0.5*NP.dot(NP.dot(md,icov),dm)
    thedet=NP.linalg.det(cov)
    denom=( math.pow((2*NP.pi),float(size)/2) * math.pow(thedet,1.0/2) )
    return num-math.log(denom)

def sampleObservation(fvars,f,gamma):
    fb = getbvars(fvars,f)
    fbo=map(lambda fv: NP.random.normal(fv,gamma), fb)
    return fbo
    

#computes the mean of an unweighted array of States
def sampleMeanUnweighted(samples):
    i=0
    avgstate = samples[0].get_copy()
    for s in samples[1:]:
        avgstate +=  s
    avgstate = avgstate/len(samples)
    return avgstate

#computes the mean of a weighted array of states
def sampleMean(samples):

    sumw = reduce(lambda x,y: x+float(y.weight), samples, 0.0)
    avgstate = samples[0].apply_weight()
    if sumw:
        for s in samples[1:]:
            avgstate = avgstate+s.apply_weight()
        avgstate = avgstate/sumw
    else:
        avgstate=sampleMeanUnweighted(samples)
    return avgstate


def fsampleMean(fsamples):
    avgf = fsamples[0]
    for f in fsamples[1:]:
        avgf = avgf+f
    avgf = avgf/len(fsamples)
    
    return avgf



def fdist(oval,fbehv):
    cd=[float(fbehv["e"]),float(fbehv["p"]),float(fbehv["a"])]
    dd=NP.array(oval)-NP.array(cd)
    return NP.dot(dd,dd)

def findNearestBehaviour(oval,fbeh):
    start=True
    mind=0
    for f in fbeh:
        fd = fdist(oval,fbeh[f])
        if start or fd<mind:
            mind=fd
            bestb=f
            start=False
    return bestb

#inserts ld=(label,dist) into list of such pairs
#up to a maximum of N, sorted by increasing dist
def insert_into(lds,ld,N):
    i=0
    while i<N and ld[1]>lds[i][1]:
        i += 1
    while i<N:
       keepit=lds[i]
       lds[i]=ld
       ld=keepit
       i += 1
    return lds

#find N nearest behaviours in database to oval
def findNearestBehaviours(oval,fbeh,N):
    mind=0
    bestbs=zip(N*[''],N*[1000.0])
    for f in fbeh:
        fd = fdist(oval,fbeh[f])
        bestbs=insert_into(bestbs,(f,fd),N)
    return map(lambda x: x[0],bestbs)
        
def findNearestIdentity(oval,fid):
    start=True
    mind=0
    for f in fid:
        fd = fdist(oval,fid[f])
        if start or fd<mind:
            mind=fd
            bestb=f
            start=False
    return bestb


def drawSamples(samples,N=-1):
    if N==-1:
        N=len(samples)

    weights=map(lambda x: x.weight,samples)
    cumweights=NP.cumsum(weights);
    cumw=cumweights/cumweights[-1]
    #don't actually need to do this - can just selecti
    #one random point in the first interval 1..(1/N)
    #and then uniform multiples of 1/N from there
    rsn=NP.random.random_sample((N,))
    rsns=NP.sort(rsn)

    i=0
    new_samples=[]
    for rs in rsns:
        while cumw[i]<rs:
            i+=1
        thesample=samples[i].get_copy()
        thesample.weight=1.0
        new_samples.append(thesample)
    return new_samples



#print all samples     
def printSamples(samples):
    map(lambda x: x.print_val(),samples)

def printSamplesFW(samples):
    map(lambda x: x.print_fw(),samples)

#only print samples with a weight greater than heavy
def printHeavySamples(samples,heavy):
    map(lambda x: x.print_val(), filter(lambda x: x.weight>heavy, samples))



#---------------------------------------------------------------------------------------------------------------------------
#class to describe the state - including fundamental and transient sentiments and x
#this will be used for sampling, so includes a weight
#---------------------------------------------------------------------------------------------------------------------------
class State(object):
    def __init__(self,f,tau,x,weight):
        self.f=f
        self.tau=tau
        self.x=x
        self.weight=weight

    def __add__(self,other):
        return State(self.f+other.f,self.tau+other.tau,self.x+other.x,self.weight+other.weight)
    def __mul__(self,other):
        return State(self.f*other.f,self.tau*other.tau,self.x*other.x,self.weight+other.weight)
    def __div__(self,other):
        #hmmm - how can this work? 
        return State(NP.divide(self.f,other),NP.divide(self.tau,other),NP.divide(self.x,other),NP.divide(self.weight,other))
    def __iadd__(self,other):
        self.f=NP.add(self.f,other.f)
        self.tau=NP.add(self.tau,other.tau)
        self.x=NP.add(self.x,other.x)
        self.weight=NP.add(self.weight,other.weight)
        return self

    def get_copy(self):
        return State(self.f,self.tau,self.x,self.weight)

    def apply_weight(self):
        return State(NP.dot(self.f,self.weight),NP.dot(self.tau,self.weight),NP.dot(self.x,self.weight),self.weight)

    #add some noise to samples ("roughening" - see Gordon et al. 1993 or Chapt. 10 of Nando's book)
    def roughen(self,noiseVector):
        nois=(2*NP.random.random((1,9))-1.0)*noiseVector
        self.f = self.f + nois.flatten(1)
        #nois=(2*NP.random.random((1,9))-1.0)*noiseVector
        #self.tau = self.tau + nois.flatten(1)
    

    def print_val(self):
        print 50*"~"
        self.print_fw()
        print "tau :",self.tau
        print "x :",self.x
        print 50*"~"

    def print_fw(self):
        print "f :",self.f
        print "weight :",self.weight

#---------------------------------------------------------------------------------------------------------------------------
#The main Agent Class
#this can be used directly or subclassed (see TutoringAgent below)
#This basic version has a single continuous "x" state that does nothing at all
#is just a placeholder really that could be removed but helps to see what to do with x
#---------------------------------------------------------------------------------------------------------------------------
class Agent(object):
    
    def __init__(self,*args,**kwargs):
        #parameter of the deflection
        self.alpha_value=kwargs.get("alpha_value",1.0)

        #parameter of the fundamentals identity prior
        #this first one governs how much variance we expect in the identities - should normally be small
        #if  we don't expect identities to change much
        #AAB6D
        #beta_value_agent should be set to 1e-20 if we are using a 6-D aab vector 
        #the agent is in charge of setting its own identity at each iteration
        self.beta_value_agent=kwargs.get("beta_value_agent",0.01)
        self.beta_value_client=kwargs.get("beta_value_client",0.01)
        self.beta_value_agent_init=kwargs.get("beta_value_agent_init",0.01)
        self.beta_value_client_init=kwargs.get("beta_value_client_init",0.01)

        #this governs the variance in behaviours - should normally be large in fact infinite, but this causes all sorts of
        #problems so we make it really really big
        self.beta_valueb_client=kwargs.get("beta_valueb_client",1e+20)
        self.beta_valueb_client_init=kwargs.get("beta_valueb_client_init",1.0)
        self.beta_valueb_agent=kwargs.get("beta_valueb_agent",1e-20)
        self.fifname=kwargs.get("identities_file","fidentities.dat")

        #parameter of the observation function - should be related to o_hitrate below
        self.gamma_value=kwargs.get("gamma_value",1.0)

        #discount factor!
        self.discount_factor=kwargs.get("discount_factor",0.9)

        #number of samples to use
        self.N=kwargs.get("N",300)


        self.client_gender=kwargs.get("client_gender","male")
        self.agent_gender=kwargs.get("agent_gender","male")

        self.agent_rough=kwargs.get("agent_rough",0.0)
        self.client_rough=kwargs.get("client_rough",0.0)


        #some defaults - 
        self.x_avg=0
        self.deflection_avg=0

        #samples
        self.samples=[]
        self.init_covariances()
        self.init_sentiment_dictionaries(self.fifname)
        self.init_HC()

    def print_params(self):
        print "alpha_value: ",self.alpha_value
        print "beta_value_agent: ",self.beta_value_agent
        print "beta_value_client: ",self.beta_value_client
        print "beta_value_client_init: ",self.beta_value_client_init
        print "beta_value_agent_init: ",self.beta_value_agent_init
        print "beta_valueb_client: ",self.beta_valueb_client
        print "beta_valueb_agent: ",self.beta_valueb_agent
        print "gamma_value: ",self.gamma_value
        print "discount_factor: ",self.discount_factor
        print "N: ",self.N
        print "client_gender: ",self.client_gender
        print "agent_gender: ",self.agent_gender
        print "agent_rough: ",self.agent_rough
        print "client_rough: ",self.client_rough

        
    def init_covariances(self):
        #set up the covariance matrices
        e3=NP.eye(3)
        z3=NP.zeros((3,3))
        betabc2=self.beta_valueb_client**2
        betabci=self.beta_valueb_client_init**2
        betaba2=self.beta_valueb_agent**2
        betaa2=self.beta_value_agent**2
        betac2=self.beta_value_client**2
        betaci=self.beta_value_client_init**2
        betaai=self.beta_value_agent_init**2
        sigfa1 = NP.vstack([NP.hstack([e3, z3, z3]),NP.hstack([z3, z3, z3]),NP.hstack([z3, z3, z3])])*betaa2
        sigfai = NP.vstack([NP.hstack([e3, z3, z3]),NP.hstack([z3, z3, z3]),NP.hstack([z3, z3, z3])])*betaai
        sigfc1 = NP.vstack([NP.hstack([z3, z3, z3]),NP.hstack([z3, z3, z3]),NP.hstack([z3, z3, e3])])*betac2
        sigfci = NP.vstack([NP.hstack([z3, z3, z3]),NP.hstack([z3, z3, z3]),NP.hstack([z3, z3, e3])])*betaci
        sigf2bc = NP.vstack([NP.hstack([z3, z3, z3]),NP.hstack([z3, e3, z3]),NP.hstack([z3, z3, z3])])*betabc2
        sigf2bci = NP.vstack([NP.hstack([z3, z3, z3]),NP.hstack([z3, e3, z3]),NP.hstack([z3, z3, z3])])*betabci
        sigf2ba = NP.vstack([NP.hstack([z3, z3, z3]),NP.hstack([z3, e3, z3]),NP.hstack([z3, z3, z3])])*betaba2

        sigfb = e3*betabc2
        isigfb=e3/betabc2

        self.sigf={}
        self.isigf={}
        self.sigf_unconstrained_b={}
        self.isigf_unconstrained_b={}


        
        #sigfa1: betaa2 : beta_value_agent^2 over agent id  (e.g. 0.001)
        #sigfc1: betac2 : beta_value_client^2 over client id (e.g. 0.001)
        #sigf2bc: betabc2 : beta_valueb_client^2 over behaviour  (e.g. 1e+20 or unconstrained)
        self.sigf["client"]=sigfa1+sigfc1+sigf2bc
        self.isigf["client"] = NP.linalg.inv(self.sigf["client"])

        #sigf2ba: betaba2 : beta_valueb_agent^2 over behaviour (e.g. 1e-20 or constrained to be agent action)
        self.sigf["agent"]=sigfa1+sigfc1+sigf2ba
        self.isigf["agent"] = NP.linalg.inv(self.sigf["agent"])

        
        #for agent with unconstrained behaviour
        self.sigf_unconstrained_b["agent"]=sigfa1+sigfc1+sigf2bc
        self.isigf_unconstrained_b["agent"] = NP.linalg.inv(self.sigf_unconstrained_b["agent"])

        #for client with unconstrained beahaviour (is the same as above right now, but could be different)
        self.sigf_unconstrained_b["client"]=sigfa1+sigfc1+sigf2bc
        self.isigf_unconstrained_b["client"] = NP.linalg.inv(self.sigf_unconstrained_b["client"])


        self.sigfi=sigfai+sigfci+sigf2bci
        self.isigfi = NP.linalg.inv(self.sigfi)


        alpha2=self.alpha_value**2
        self.siga=NP.eye(9,9)*alpha2
        self.isiga=NP.eye(9,9)/alpha2
        self.sigab = e3*alpha2
        self.isigab = e3/alpha2

        self.gamma_value2=self.gamma_value*self.gamma_value


    def init_sentiment_dictionaries(self,fifname):

        self.fidentities_client=readSentiments(fifname,self.client_gender)
        self.fidentities_agent=readSentiments(fifname,self.agent_gender)

        if self.agent_gender=="female":
            self.tdyn=tdyn_female
        else:
            self.tdyn=tdyn_male

        #set up the swapped tvar matrices and G,M matrices
        self.G={}
        self.M={}
        self.GM={}
        for tturn in ("agent","client"):
            self.G[tturn]={}
            self.M[tturn]={}
            self.GM[tturn]={}

        for tvar in tvars:
            self.G["agent"][tvar]=map(lambda xx: xx[0], self.tdyn)
            self.M["agent"][tvar]=map(lambda xx: xx[1][tvars.index(tvar)], self.tdyn)
            self.G["client"][tvar]=map(lambda xx: swapACVars(xx[0]), self.tdyn)
            self.M["client"][tvar]=map(lambda xx: xx[1][tvars.index(swapACVar(tvar))], self.tdyn)
            self.GM["agent"][tvar]=zip(self.G["agent"][tvar],self.M["agent"][tvar])
            self.GM["client"][tvar]=zip(self.G["client"][tvar],self.M["client"][tvar])

        

    def init_HC(self):
        #now we construct the functions that will compute the H and C matrices
        #which is 9x3 with the jth column giving the terms of M't that contain
        #B_{d_a(j)} and C is a 9x1 vector that gives the terms of M't that contain
        #no B variable.  M and C are functions of tau and x' (the turn)
        #the turn (x') is necessary as we have to swap all A and C variables
        #if the current turn (x) is the client (so x'=agent)
        Hagent={}
        for j in ["Tbe","Tbp","Tba"]:
            Hagent[j]=filter(lambda xx: has_variable(xx[0],j),self.tdyn)
    
        Cagent=filter(lambda xx: not has_variable(xx[0],"Tb"), self.tdyn)

        Hclient={}
        for j in ["Tbe","Tbp","Tba"]:
            Hclient[j]=swapAC(filter(lambda xx: has_variable(xx[0],j),self.tdyn))
        Cclient=swapAC(filter(lambda xx: not has_variable(xx[0],"Tb"), self.tdyn))

        self.H={}
        self.H["client"]=Hclient
        self.H["agent"]=Hagent

        self.C={}
        self.C["client"]=Cclient
        self.C["agent"]=Cagent

    def initialise_x(self,initx):
        return initx

    def initialise(self,f,initx):
        #draw a set of samples from the initial distributions
        self.samples=[]
        for i in range(self.N):
            tmpfv=NP.random.multivariate_normal(f,self.sigfi)
            tmpix=self.initialise_x(initx)
            xx=State(tmpfv,tmpfv,tmpix,1.0)
            self.samples.append(xx)
        avgs=sampleMean(self.samples)
        self.x_avg=avgs.x
        return avgs

    def initialise_array(self,f,prop,initx=None):
        #draw a set of samples from an array of initial distributions with weights given by prop
        #the set is drawn according to prop so f can be of any size greater than prop
        self.samples=[]
        nums=NP.random.multinomial(self.N, prop)
        j=0
        self.init_f=f
        self.init_prop=prop
        self.init_initx=initx
        for n in nums:
            for i in range(n):
                tmpfv=NP.random.multivariate_normal(f[j],self.sigfi)
                tmptv=NP.random.multivariate_normal(f[j],self.sigfi)
                tmpix=self.initialise_x(initx)
                xx=State(tmpfv,tmpfv,tmpix,1.0)
                self.samples.append(xx)
            j += 1
        avgs=sampleMean(self.samples)
        self.x_avg=avgs.x
        return avgs

    def initialise_array_take_action(self,f,prop,initx=None):
        avgs = self.initialise_array(f,prop,initx)
        return self.update_agent(avgs)
        
    
    def sample_next_f(self,state,N=100,turn="agent",meth="full"):
        #draw a set of samples over the next fundamental sentiment
        fvsamples=[]
        for x in range(N):
            (tmpfv,wgt)=sampleFvar(fvars,tvars,self.H,self.C,self.isiga,self.isigf_unconstrained_b,state.tau,state.f,turn,[],[],meth)
            tmpfv=tmpfv.transpose()
            fvsamples.append(tmpfv)
        avgf=fsampleMean(fvsamples)
        return avgf


    #default value simply returns the input x
    def sampleXvar(self,f,tau,x,turn,aab,paab=None):
        return x

    def rollout(self,fvars,tvars,state,turn,depth,meth="full"):
        N=20
        width=1.0
        epsilon = 0.001

        if math.pow(self.discount_factor,depth) < epsilon:
            return 0

        #need to check if turn="client" here and do something different if it is
    
        #rollout policy selects near the optimal myopic action to take
        (fsample,wgt)=sampleFvar(fvars,tvars,self.H,self.C,self.isiga,self.isigf_unconstrained_b,state.tau,state.f,turn,[],[],meth)
        fsample=fsample.transpose()
        a=map(lambda x: float(x), fsample[3:6])
        #simulate next state given that action
        sample=self.sampleNext(fvars,tvars,state,turn,a,[],meth)

        #get reward of that - turn should switch here I guess??
        return self.reward(sample)+self.discount_factor*self.rollout(fvars,tvars,sample,turn,depth+1,meth)


                                       
    def reward(self,state):
        # a generic deflection-based reward that can be removed eventually
        # this can be removed??
        fsample=state.f.transpose()
        reward = -1.0*NP.dot(state.f-state.tau,state.f-state.tau)
        return reward


    #draw a next sample from a currrent one "state" by sampling f, tau and x in that order
    def sampleNext(self,fvars,tvars,state,turn,aab,observ,paab=None,meth="full"):

        #these are generic methods that can be outside the class
        (fsample,wgt)=sampleFvar(fvars,tvars,self.H,self.C,self.isiga,self.isigf,state.tau,state.f,turn,aab,observ,meth)

        #old way:
        #tsample=sampleTvarsOld(tvars,self.tdyn,state.tau,fsample,turn)
        #new way (faster)
        tsample=sampleTvars(tvars,self.GM,state.tau,fsample,turn)
        #this is class dependent
        xsample=self.sampleXvar(fsample,tsample,state.x,turn,aab,paab)
        return State(fsample,tsample,xsample,wgt)


    #this must be in the class
    def evalSampleXvar(self,sample,turn,bact,xobs):
        #default no weight
        return 1.0
            

    #add some noise to samples ("roughening" - see Gordon et al. 1993 or Chapt. 10 of Nando's book)
    def roughenSamples(self,samples):
        noiseVector=NP.concatenate((NP.ones((3,1))*self.agent_rough,NP.zeros((3,1)),NP.ones((3,1))*self.client_rough)).flatten(1)
        map(lambda x: x.roughen(noiseVector),samples)

    #the main SMC update step. Given an affective action aab and a propositional action paab,
    #and an affective observation observ and an  observation xobserv, update the samples
    #by first drawing an unweighted set, propagating those forward, computing the weights
    #and finally computing the deflection
    def propagate_forward(self,turn,aab,observ,xobserv=[],paab=None,meth="full",verb=False):

        if False and verb:
            print 100*"!","samples before resample:"
            printSamples(self.samples)
        
        #resample to get an unweighted set
        new_samples=drawSamples(self.samples)

        #add "jitter" here instead of the proposal idea...
        #possibly just add the jitter to the client identity
        if meth=="rough":
            self.roughenSamples(new_samples)

        if verb:
            print 100*"!","unweighted set:"
            printSamplesFW(new_samples)

        #propagate forward
        self.samples=[]
        totweight=0.0
        for sample in new_samples:
            newsample = self.sampleNext(fvars,tvars,sample,turn,aab,observ,paab,meth)
            theweight=evalSampleFvar(tvars,self.tdyn,newsample,self.gamma_value2,turn,observ)
            newweight = newsample.weight+theweight
            newsample.weight=math.exp(newweight)
            totweight += newsample.weight
            self.samples.append(newsample)
        
        if totweight<1e-20:
            #reset 
            print 100*"r","eset"
            tmp_avgs=self.initialise_array(self.init_f,self.init_prop,self.init_initx)




        if False and verb:
            print 100*"!","new samples:"
            printHeavySamples(self.samples,0)

        avgstate=sampleMean(self.samples)
        self.deflection_avg=NP.dot(avgstate.tau-avgstate.f,avgstate.tau-avgstate.f)
        self.x_avg=avgstate.x
        
        return avgstate


    #primary update function - takes a client action and returns an agent action
    #actually updates the model twice: once for client, once for agent
    def update(self,observ,xobserv=[],meth="full",verbose=False):
        #propagate for client
        #returns the average fundamental sentiments
        avgs=self.propagate_forward("client",[],observ,xobserv,None,meth,verbose)
        print "client average sentiments: "
        avgs.print_val()
        return self.update_agent(avgs,meth,verbose)

    #default when there is no propositional action
    def get_prop_action(self):
        return None

    def get_next_action(self,state,turn,paab=None,meth="full"):
        # this should be
        # consult policy using current tau and f --> aab to take (EPA vector)
        # alternatively, we may have a direct mapping from string+difficulty -> EPA
        # which we can search in reverse to find the best string+difficulty given the EPA value of aab
        # given by the policy
        # for now, we just ask manually what the label is
        #print "rollout value from this state is : ",self.rollout(fvars,tvars,state,turn,0,meth="full")
        #ba=findBestAction(f,tau,3,fvars,tvars,tdyn,alpha_value,beta_value,gamma_value)

        #right now, am using these as action selectors
        avgf=self.sample_next_f(state,100,turn,meth)
        aab=map(lambda x: float(x), avgf[3:6])
        return aab
    

    def update_agent(self,avgs,meth="full",verbose=False):

	propositional_action = self.get_prop_action()
	
        #this needs to be replaced with the full Monte-Carlo solution once developed
        affect_action=self.get_next_action(avgs,"agent",propositional_action,meth=meth)   
        #propagate for agent
        print "agent action: ",affect_action
        avgs=self.propagate_forward("agent",affect_action,[],[],propositional_action,meth,verbose)
        print "agent average sentiments: "
        avgs.print_val()
        return (affect_action,propositional_action)

    def compute_deflection(self):
        self.deflection=0
        sumw=0
        for sample in self.samples:
            self.deflection += sample.weight*NP.dot(sample.f-sample.tau,sample.f-sample.tau)
            sumw += sample.weight
        self.deflection /= sumw
        return self.deflection

    def getAverageSentiments(self):
        avgstate=sampleMean(self.samples)
        return avgstate
    
    def get_avg_ids(self,f):
        aid=findNearestIdentity(map(lambda x: x[1],filter(lambda x: x[0][1]=="a",zip(fvars,f))),self.fidentities_agent)
        cid=findNearestIdentity(map(lambda x: x[1],filter(lambda x: x[0][1]=="c",zip(fvars,f))),self.fidentities_client)
        return (aid,cid)

        #find identities across all samples
    def get_all_ids(self):
        print "extracting ids from samples ..."
        agent_ids=map(lambda z: findNearestIdentity(map(lambda x: x[1],filter(lambda y: y[0][1]=="a",zip(fvars,z.f))),self.fidentities_agent),self.samples)
        client_ids=map(lambda z: findNearestIdentity(map(lambda x: x[1],filter(lambda y: y[0][1]=="c",zip(fvars,z.f))),self.fidentities_client),self.samples)

        cnt_ag={}
        for aid in agent_ids:
            if cnt_ag.has_key(aid):
                cnt_ag[aid] += 1
            else:
                cnt_ag[aid] = 0 
        cnt_cl={}
        for cid in client_ids:
            if cnt_cl.has_key(cid):
                cnt_cl[cid] += 1
            else:
                cnt_cl[cid] = 0 

        cnt_ags = sorted(cnt_ag.items(), key=lambda x: x[1],reverse=True)
        cnt_cls=sorted(cnt_cl.items(), key=lambda x: x[1],reverse=True)
        return (cnt_ags,cnt_cls)

        

